; Butterfly BASIC - a BASIC interpreter for the MSP430
; Copyright (c) 2003, 2004 Paul Curtis
; email: plc@rowley.co.uk

; Butterfly Basic Register Usage
; ------------------------------
; R4 - current token taken from tokenized program image
; R5 - pointer to start of line currently executing
; R6 - pointer to current token within executing line
; R11 - for use exclusively by statement parsers
; R12/R13 - evaluation next on stack when entering binary operator routines
; R14/R15 - evaluation top of stack

#define tp  r6
#define lp  r5

#define BREAK 0x01
#define TIMER 0x02

; Hardware multiplier registers
#if 0
MPY     equ     0x130
MPYS    equ     0x132
MAC     equ     0x134
MACS    equ     0x136
OP2     equ     0x138
RESLO   equ     0x13a
RESHI   equ     0x13c
SUMEXT  equ     0x13e
#endif

#define COMMAND_TOKENS 0
#define UNARY_TOKENS   1
#define MISC_TOKENS    2
#define BINARY_TOKENS  3
                                 
#define END_OF_RAM 0xa00

        INCLUDELIB "c_sf"

; Keywords
#define T(SYM, STR, FUNC, PREC) _T SYM

_T	macro	X
X	equ	symval
symval	set	symval+1
	endm

#define USER_TOKENS "user_tokens.inc"

symval              set     $80
FIRST_COMMAND_TOKEN equ symval
#undef PART
#define PART COMMAND_TOKENS
#include "../tokens.inc"
#include "user_tokens.inc"
LAST_COMMAND_TOKEN  equ symval-1
FIRST_UNARY_TOKEN   equ symval
#undef PART
#define PART UNARY_TOKENS
#include "../tokens.inc"
#include "user_tokens.inc"
LAST_UNARY_TOKEN    equ symval-1
FIRST_MISC_TOKEN    equ symval
#undef PART
#define PART MISC_TOKENS
#include "../tokens.inc"
#include "user_tokens.inc"
LAST_MISC_TOKEN     equ symval-1
FIRST_BINARY_TOKEN  equ symval
#undef PART
#define PART BINARY_TOKENS
#include "../tokens.inc"
#include "user_tokens.inc"
LAST_BINARY_TOKEN   equ symval-1
LAST_TOKEN          equ symval

; Token command table
#undef T
#define T(SYM, STR, FUNC, PREC) _C FUNC

_C	macro	X
	dw      X
	endm

        .code
cmdtab:
#undef PART
#define PART COMMAND_TOKENS
#include "../tokens.inc"
#include "user_tokens.inc"

binarytab:
#undef PART
#define PART BINARY_TOKENS
#include "../tokens.inc"
#include "user_tokens.inc"

unarytab:
#undef PART
#define PART UNARY_TOKENS
#include "../tokens.inc"
#include "user_tokens.inc"

; Binary operator precedence table
#undef T
#define T(SYM, STR, FUNC, PREC) _P PREC

_P	macro	X
	db      X
	endm

        .code
operator_precedence:
        dc.b    0
#undef PART
#define PART BINARY_TOKENS
#include "../tokens.inc"
#include "user_tokens.inc"

; Keyword table
#undef T
#define T(SYM, STR, FUNC, PREC) _K STR

_K	macro	X
	db      X, 0
	endm

        .code
kwtab:
#undef PART
#define PART COMMAND_TOKENS
#include "../tokens.inc"
#include "user_tokens.inc"
#undef PART
#define PART UNARY_TOKENS
#include "../tokens.inc"
#include "user_tokens.inc"
#undef PART
#define PART MISC_TOKENS
#include "../tokens.inc"
#include "user_tokens.inc"
#undef PART
#define PART BINARY_TOKENS
#include "../tokens.inc"
#include "user_tokens.inc"


; Variable storage
        .bss
current_channel ds.b    1
trace_channel   ds.b    1
trace_flags     ds.b    1
        even
vars            ds.w    26*2                 ; A to Z is 104 bytes
arrays          ds.w    26*2                 ; A() to Z() is 104 bytes; 2 bytes for address, 2 for dimension
ramtop          ds.w    1
flashtop        ds.w    1
break_tp        ds.w    1                    ; break vector
break_lp        ds.w    1                    ; start of line of last ON BREAK
data_lp         ds.w    1                    ; pointer to start of line of current DATA statement
data_tp         ds.w    1                    ; pointer for DATA statements on READ
cont_tp         ds.w    1
cont_lp         ds.w    1
flags           ds.w    1
rndseed         ds.w    2
auto_line       ds.w    1                    ; line number for AUTO
auto_inc        ds.w    1                    ; line increment for AUTO
convert_buffer  ds.b    16

; This points to the (logical) first line of code in a user's program.
; As the program is edited, program_start will move to point to the
; lowest-numbered line number.
program_start   ds.w    1

        .code
	even

reset::
; Zero the whole of UDATA
        mov.w   #SFB(UDATA0), r15
        mov.w   #SFE(UDATA0)-SFB(UDATA0), r13
reset_10:
        mov.b   #0, @r15
	add.w   #1, r15
	sub.w   #1, r13
	jnz     reset_10

; Entry point for cold start.
        mov.w   #WDTPW+WDTHOLD, &WDTCTL      ; Turn off watchdog
        mov.w   #END_OF_RAM, sp              ; Initialise stack
        call    #hw_init                     ; Initialise hardware environment
        mov.b   #1, &current_channel         ; Default channel for output and tracing is #1
        mov.b   #1, &trace_channel
        mov.b   #0, &trace_flags             ; Start by not tracing

; Sign on.
        mov.w   #copyright_message, r15
        call    #print_message

; Prepare memory and show how much there is to the user.
        call    #new
        call    #memory

; Initialise PRNG.
        mov.w   #25374, &rndseed[0]
        mov.w   #63398, &rndseed[1]

; Come out of AUTO mode and go to ready.
no_auto:
        mov.w   #0, &auto_line               ; drop out of AUTO mode

; Entry point for main BASIC command entry loop
ready:
        mov.w   #1, r15                      ; Select the default channel
        call    #select_channel
        mov.w   #ready_message, r15          ; Indicate we're ready to accept a command
        call    #print_message
	mov.w   &ramtop, r14                 ; prepare to read string into ramtop
	cmp.w	#0, &auto_line               ; under control of AUTO?
	jz      ready_20
	mov.w   &auto_line, r14              ; load line number
	add.w   &auto_inc, &auto_line        ; update line number for next time round
	jnc     ready_05                     ; did the line number wrap?
        mov.w   #0, &auto_line               ; ...yes, come out of AUTO
ready_05:
	mov.w   #0, r15                      ; zero extend for conversion
	call    #print_integer               ; print it and also put it in the conversion buffer
	mov.w   #convert_buffer, r15         ; prepare to copy from convert buffer...
	mov.w   &ramtop, r14                 ; ...to ramtop
ready_10:
	mov.b   @r15+, @r14                  ; copy byte
	add.w   #1, r14                      ; no auto increment on destination
	cmp.b   #0, -1(r15)                  ; end of string copied?
	jnz     ready_10                     ; no, continue copying...
	mov.b   #' ', r15                    ; prepare to
	call    #print_char                  ; print a space
	mov.b   r15, -1(r14)                 ; and put it in the buffer
ready_20:
	call    #get_string_00               ; read string to memory at R14
        mov.w   #0, lp                       ; we're not executing a line stored in flash, so errors don't report a line number

; Skip initial spaces
        mov.w   &ramtop, tp
interpreter_00:
        mov.b   @tp+, r4                     ; Get character from buffer         
        cmp.b   #' ', r4                     ; if it's a space, ignore it
        jz      interpreter_00
        sub.w   #1, tp                       ; we've gone one too far with autoincrement, so adjust
        cmp.b   #0, r4                       ; if nothing on the entered line...
        jz      no_auto                      ; ...just be silent, drop AUTO, and get another line

; Figure out whether this is an immediate command or a numbered line
        call    #is_digit
        jc      interpreter_20               ; if carry set, then a line with line number has been entered

; No line nunber, so this is an immediate command
	mov.w   #0, &auto_line               ; come out of AUTO
        call    #tokenize                    ; tokenize put
        mov.w   &ramtop, tp                  ; start interpreting from the tokenized buffer
        jmp     chrget                       ; and execute

; Deal with a numbered line
interpreter_20:
        add.w   #1, tp                       ; point to next character
        call    #read_int                    ; read the line number into r15:r14...
        push.w  r14                          ; ...but only use the lower 16 bits
        sub.w   #1, tp                       ; we've run one too far on the line number
        mov.w   tp, r9                       ; save token pointer
        call    #tokenize                    ; tokenize the line
        pop.w   r14                          ; restore the line number
        mov.w   #program_start, r11          ; pointer to previous line

; Come out of AUTO mode if no line
        cmp.b   #0, @r9                      ; empty line?
	jnz     interpreter_30               ; ...no, process the line
        cmp.w   #0, &auto_line               ; in AUTO mode?
        jnz     no_auto                      ; ...yes, drop AUTO mode and go to ready mode

; Copy line into flash
interpreter_30:
        mov.w   &flashtop, r10
        mov.w   r10, r7
        mov.w   r14, r4
        add.w   #2, r10
        call    #flash_write_word
        add.w   #2, r10
copy_line:
        mov.b   @r9+, r4
        call    #flash_write_byte
        add.w   #1, r10
        cmp.w   #0, r4
        jnz     copy_line
        add.w   #1, r10
        bic.w   #1, r10
        mov.w   r10, &flashtop

; Find line number in r14
        mov.w   #program_start, r11
        mov.w   &program_start, r15
        cmp.w   #0, r15
        jz      insert_new_line
        jmp     x_find_line_10
x_find_line_00:
        mov.w   r15, r11
        mov.w   @r15, r15
        cmp.w   #0, r15
        jz      insert_new_line
x_find_line_10:
        cmp.w   2(r15), r14
        jz      replace_existing_line
        jc      x_find_line_00

; Insert line at r15.
insert_new_line:
        cmp.b   #0, 4(r7)                    ; Nothing to insert, i.e. deleting a non-existent line?
        jz      ready                        ; Yep, nothing to do.

; Point new line's foward pointer to point at next line.
        mov.w   r7, r10
        mov.w   @r11, r4
        call    #flash_write_word

; Update previous line's forward pointer to point here.
        mov.w   r7, r4
        mov.w   r11, r10
        call    #flash_write_word
        jmp     ready

replace_existing_line:
; Delete line entirely?
        cmp.b   #0, 4(r7)
        jz      delete_line

; Update previous line's forward pointer to point here.
        mov.w   @r15, r4                     ; data to write
        mov.w   r7, r10                      ; address to write at
        call    #flash_write_word

; Point previous line's forward pointer to point at next line.
        mov.w   r11, r10
        mov.w   r7, r4
        call    #flash_write_word
        jmp     ready

; Delete line
delete_line:
        mov.w   r11, r10
        mov.w   @r15, r4
        call    #flash_write_word
        jmp     ready

; NEW
x_new:
        call    #new
        mov.w   #END_OF_RAM, sp              ; FIXME!
        jmp     ready

; Erase whole of flash
new:
        mov.w   #BREAK, &flags               ; BREAK ON
        call    #clear                       ; zero variables
        mov.w   #user_flash_begin, r10       ; erase flash pages
x_new_00:
        call    #flash_erase_page
#ifndef SIMULATOR
        add.w   #0x200, r10
        cmp.w   #0xfe00, r10
        jnz     x_new_00
#endif

        mov.w   #user_flash_begin+4, &flashtop
        mov.w   #0, r4
        mov.w   #user_flash_begin, r10
        call    #flash_write_word
        add.w   #2, r10
        call    #flash_write_word
        mov.w   #0, &program_start

        mov.w   #0, lp
        mov.w   #ENDOF(UDATA0), &ramtop
        ret

; RUN
x_run:
; Compute start line number
        mov.w   #0, r14                      ; prepare to run from the start
	cmp.b   #0, r4                       ; anything after RUN?
	jz      x_run_05                     ; ...no, skip parsing the line number
	call    #read_int                    ; get line number to start from

; Re-initialise interpreter
x_run_05:
        mov.w   #BREAK, &flags               ; allow break in processing
        mov.w   #END_OF_RAM, sp              ; initialise stack pointer, clearing stack
        mov.w   #ENDOF(UDATA0), &ramtop      ; initialise TOP
        mov.w   #0, &break_tp                ; remove ON BREAK handler
        mov.w   #0, &break_lp
	call    #clear                       ; zero variables on run

; Before running, check structural integrity of program
        mov.w   &program_start, lp           ; get start of program into lp
        cmp.w   #0, lp                       ; any program in flash?
        jz      ready                        ; ...no, nothing to run
        mov.w   lp, tp                       ; point tp to the start of the first line
        add.w   #4, tp
        mov.b   @tp+, r4                     ; get first token from program
        mov.w   #LAST_TOKEN, r11             ; special token for find_control_match
        call    #find_control_match          ; ensure loops match
        jc      x_run_10                     ; if controls match, that's great
        cmp.w   #T_END, r4                   ; END on its own?
        jnz     syntax_error                 ; ...no, then it's a syntax error
x_run_10:
        mov.w   &program_start, lp           ; get start of program
        mov.w   lp, tp                       ; point to start of line
        push.w  #T_END                       ; delimit nothing on stack (no loops, no RESUME)
	cmp.w   #0, r14                      ; run from start?
        jz      next_line                    ; ...yes, get first line
	jmp     x_goto_00

term:
        cmp.b   #':', r4                     ; statement separator?
        jeq     chrget                       ; ...yes, start interpreting next statement
        cmp.b   #0, r4                       ; end of line?
        jeq     term_10                      ; ...yes, skip to next line
        cmp.b   #T_TICK, r4                  ; comment?
        jz      x_rem                        ; ...yes, treat as REM
	jmp     syntax_error                 ; ...no, can't be anything other than an error
term_10:
        cmp.w   #0, lp                       ; at end of line, interpreting in immediate mode?
        jeq     ready                        ; ...yes, go back to ready loop
skip_line:
        mov.w   @lp, tp                      ; point tp to the next line to interpret by following forward link
        mov.w   tp, lp                       ; save start of line in lp
        cmp.w   #0, tp                       ; run off end of program?
        jz      ready                        ; ...yes, return to ready loop
next_line:
        add.w   #4, tp                       ; skip over line number and link
chrget:
        mov.b   @tp+, r4
chrgot:

; Break-in processing.
        bit.w   #BREAK, &flags               ; break-ins enabled?
        jz      chrgot_20                    ; ...no, carry on with life
        BRKTEST chrgot_20
        cmp.w   #0, &break_tp                ; ON BREAK set?
        jz      x_stop                       ; ...no, stop execution
        mov.w   tp, &cont_tp                 ; save current context for CONTINUE
        mov.w   lp, &cont_lp
        mov.w   &break_tp, tp                ; activate ON BREAK context
        mov.w   &break_lp, lp
        jmp     chrget                       ; start ON BREAK processing

chrgot_20:
#if 0
        bit.w   #TIMER, &flags               ; timer fired?
        jz      chrgot_30                    ; no, continue execution
        bic.w   #TIMER, &flags               ; reset timer fired flag
        cmp.w   #0, &timer_tp                ; an ON TIMER handler set?
        jz      chrgot_30                    ; no, ignore timer event
        push.w  sr                           ; save current context
        dint
        push.w  tp
        push.w  lp
        push.w  #T_RESUME                    ; flag RESUME needed
        mov.w   &timer_tp, tp                ; transfer to ON TIMER context
        mov.w   &timer_lp, lp
        jmp     chrget
chrgot_30:
#endif
        cmp.b   #0, r4
        jz      skip_line
        jge     chrgot_10

; Evaluate statement
next_00:
        cmp.w   #LAST_COMMAND_TOKEN+1, r4
        jc      syntax_error
        add.b   r4, r4
        mov.b   r4, r11
        mov.b   @tp+, r4
        br      cmdtab(r11)

chrgot_10:
        cmp.b   #':', r4
        jz      chrget

; LET var = expr
; LET var(index) = expr
x_let:
	call	#varptr                      ; parse variable and get its address into R14
	mov.w   r14, r11                     ; save variable address
	push.w  r15                          ; save variable size
        cmp.b   #T_EQ, r4                    ; '=' follows?
        jnz     syntax_error                 ; ...no, syntax error
        call    #get_eval_expr               ; evaluate the expression
	pop.w   r12                          ; restore varibale designation
	cmp.w   #2, r12                      ; assigning 16 bits?
	jz      x_let_20                     ; ...yes
	cmp.w   #4, r12                      ; assigning 32 bits?
	jz      x_let_10                     ; ...yes
        mov.b   r14, @r11                    ; must be 8 bits then
        jmp     term                         ; ensure that statement is correctly terminated
x_let_10:
        mov.w   r15, 2(r11)                  ; store high part of 32-bit value
x_let_20:
        mov.w   r14, @r11                    ; store low part of 32-bit value or 16 bit value
        jmp     term                         ; ensure that statement is correctly terminated

; REM comment
x_rem:
        cmp.w   #0, lp                       ; Special REM in immediate mode
        jz      ready                        ; if in immediate mode, just say we're ready
        jmp     skip_line                    ; if in program mode, skip to the next line

; POKE address, byte
x_poke:
        call    #eval_expr                   ; evaluate address
        mov.w   r14, r11                     ; ...and only use the low 16 bits
        call    #check_comma                 ; check ',' present
        call    #eval_expr                   ; evaluate byte
        mov.b   r14, @r11                    ; do the poke
        jmp     term                         ; ensure that statement is correctly terminated

; DPOKE address, byte
x_dpoke:
        call    #eval_expr                   ; evaluate address
        bic.w   #0, r14                      ; ensure address is word aligned
        mov.w   r14, r11                     ; ...and only use the low 16 bits
        call    #check_comma                 ; check ',' present
        call    #eval_expr                   ; evaluate byte
        mov.w   r14, @r11                    ; do the dpoke
        jmp     term                         ; ensure that statement is correctly terminated

; CLEAR
x_clear:
        call    #clear                       ; zero variables
        jmp     term                         ; ensure that statement is correctly terminated

; Clear variables and arrays to zero and restore data pointer
clear:
        mov.w   #ENDOF(UDATA0), &ramtop      ; reset TOP
        mov.w   #0, r11                      ; Clear A through Z
x_clear_10:
        mov.w   #0, vars(r11)
	mov.w   #0, arrays(r11)
        add.w   #2, r11
        cmp.w   #26*4, r11
        jnz     x_clear_10
	mov.w   #0, &data_lp
        mov.w   #0, &data_tp                 ; restore to start of program
        ret

; RESTORE
; RESTORE line
x_restore:
        mov.w   #0, &data_tp                 ; by default, restore to start of program
	call    #end_of_statement            ; end of statement?
	jz      term                         ; ...yes, just restore to start of program
	call    #read_int                    ; get line number to restore to
        mov.w   tp, r11                      ; save where we are in the program
	call    #find_line                   ; find the line in the program, setting R15
	mov.w   r15, tp                      ; tp is always used to search the program
	mov.w   tp, &data_lp                 ; save start of line
	call    #find_data_statement         ; find the next DATA statement
	mov.w   tp, &data_tp                 ; save DATA statement pointer
	mov.w   r11, tp                      ; restore the interpreting tp
	mov.b   -1(tp), r4                   ; restore interpreting token
	jmp     term                         ; and check termination

; Find the next DATA statement after R15
find_data_statement:
        cmp.w   #0, tp                       ; fallen off end of program?
	jz      out_of_data_error            ; ...yes, then out of data
	add.w   #4, tp                       ; skip over line number and forward pointer
find_data_statement_10:
	mov.b   @tp+, r4                     ; get first token of statement
	cmp.b   #T_DATA, r4                  ; is this a DATA statement?
	jz      find_data_statement_return   ; ...yes, found it so return
	call    #skip_statement              ; ...no, skip this statament
find_data_statement_20:                      ; entry point from READ
	cmp.b   #0, r4                       ; need to skip over line number and forward pointer?
	jnz     find_data_statement_10       ; ...no, continue searching for the data statement
	mov.w   &data_lp, tp                 ; ...yes, so skip them
	mov.w   @tp, tp                      ; follow forward pointer
	mov.w   tp, &data_lp                 ; save start of line
	jmp     find_data_statement          ; skip past header bytes
find_data_statement_return:
	ret

; READ var, var...
x_read:
	mov.w   tp, r11                      ; save tp as we'll use tp to search for DATA statements
	mov.w   &data_tp, tp                 ; any data?
	cmp.w   #0, tp
	jnz     x_read_10                    ; yes, go to next data item
	mov.w   &program_start, tp           ; search from start of program
	mov.w   tp, &data_lp
	call    #find_data_statement         ; find data statement
	mov.w   tp, &data_tp
x_read_10:
        mov.b   @tp+, r4                     ; prime for read
	cmp.b   #',', r4                     ; comma?
	jz      x_read_10                    ; ...yes, just skip it
	call    #end_of_statement            ; come to end of DATA statement?
	jnz     x_read_20                    ; ...no, read value from program
        call    #find_data_statement_20      ; ...yes, find next DATA statement
	jmp     x_read_10                    ; and see if anything in this DATA statement
x_read_20:
        call    #read_int                    ; read value
	sub.w   #1, tp                       ; tp already advanced, so bring it back
	mov.w   tp, &data_tp                 ; save for next read
	mov.w   r11, tp                      ; restore interpreter's tp
	mov.b   -1(tp), r4                   ; prime token
	push.w  r15                          ; save read value on stack
	push.w  r14
	call    #varptr                      ; compute variable's address
	cmp.w   #4, r15                      ; only allow READ on word variables
	jnz     syntax_error
	pop.w   0(r14)                       ; store read value to variable
	pop.w   2(r14)
	cmp.w   #',', r4                     ; more to read?
	jnz     term                         ; ...no, terminate statement
	mov.b   @tp+, r4                     ; prime next token
	jmp     x_read                       ; and read again

; DATA ...
x_data:
	call    #skip_statement              ; skip the statement
	jmp     term                         ; continue with the following statement

; Skip the current statement, and point to the following statement
skip_statement:
	cmp.b   #0, r4                       ; end of line?
	jz	skip_statement_return        ; ...yes, we've skipped it
	cmp.b   #':', r4                     ; separator?
	jz	skip_statement_return        ; ...yes, we've skipped it
	cmp.b   #T_TICK, r4                  ; remark?
	jz      skip_statement_10            ; ...yes, skip to end of line
	cmp.b   #T_REM, r4                   ; remark?
	jz      skip_statement_10            ; ...yes, skip to end of line
	mov.b   @tp+, r4                     ; get next token in statement
	jmp     skip_statement               ; and continue searching for the end of the statement
skip_statement_10:
	mov.b   @tp+, r4                     ; get next token
	cmp.b   #0, r4                       ; end of line?
	jnz     skip_statement_10            ; ...no, still skipping remark
skip_statement_return:
        ret

; DIM var1(n1), var2(n2)
x_dim:
        call    #is_letter                   ; dimensioning a valid array name?
        jnc     syntax_error                 ; ...no, it's a syntax error
        mov.w   r4, r11                      ; save variable name for later
        mov.b   @tp+, r4                     ; '(' next?
        cmp.b   #'(', r4
        jnz     syntax_error                 ; ...no, it's a syntax error
        call    #eval_unary                  ; get number of elements
        cmp.w   #0, r15                      ; high word non-zero?
        jnz     out_of_memory_error          ; ...yes, far too much memory required
        cmp.w   #16384, r14                  ; more than 64/4 = 16K elements?
        jc      out_of_memory_error          ; ...yes, array too big so report as out of memory
	add.w   #1, r14                      ; DIM X(10) means X has 11 elements, 0 through 10
	mov.w   r14, r9                      ; save maximum index
        add.w   r14, r14                     ; compute number of 32-bit elements
        add.w   r14, r14
        mov.w   &ramtop, r15                 ; 16-bit align TOP
        add.w   #1, r15
        bic.w   #1, r15
        mov.w   r15, r10                     ; save start address of array in R10
        add.w   r14, r15                     ; compute new TOP
        mov.w   sp, r14                      ; gap between SP and TOP must be at least 32 bytes
        sub.w   #32, r14
        cmp.w   r14, r15                     ; gap too small?
        jc      out_of_memory_error          ; ...yes, report out of memory
        mov.w   r15, &ramtop                 ; ...no, set new TOP
	sub.w   #'A', r11                    ; compute index into array table
	add.w   r11, r11
	add.w   r11, r11
	mov.w   r10, arrays(r11)             ; set array start address
	mov.w   r9, arrays+2(r11)            ; set array maximum index
	cmp.b   #',', @tp                    ; dimension another array?
        jnz     term                         ; no, check termination
	add.w   #1, tp                       ; skip '.'
	mov.b   @tp+, r4                     ; get array name
	jmp     x_dim                        ; and dimension again

; WHILE cond ... END WHILE
x_while:
        push.w  tp                           ; stack interpreter's context
        push.w  lp
        push.w  #T_WHILE                     ; indicate that this is a WHILE loop
x_while_00:
        call    #eval_expr                   ; evaluate test expression
        bis.w   r14, r15
        cmp.w   #0, r15                      ; result is zero?
        jnz     term                         ; ...no, continue loop
        add.w   #6, sp                       ; ...yes, drop loop type, tp, and lp
        mov.w   #T_WEND, r11                 ; search for a matching WEND
        call    #find_control_match
        jnc     while_without_matching_end_while_error
        mov.b   @tp+, r4
        jmp     term

; WEND
x_wend:
        cmp.w   #T_WHILE, @sp                ; is control structure WHILE?
        jnz     end_while_without_while_error; ...no, report error
        mov.w   2(sp), lp                    ; ...yes, restore context to that following WHILE
        mov.w   4(sp), tp
        mov.b   -1(tp), r4
        jmp     x_while_00                   ; re-evaluate expression

; REPEAT ... UNTIL cond
x_repeat:
        push.w  tp                           ; stack interpreter's context
        push.w  lp
        push.w  #T_REPEAT                    ; indicate that this is a REPEAT loop
        jmp     term                         ; ensure that statement is correctly terminated

x_until:
        cmp.w   #T_REPEAT, @sp               ; control structure is REPEAT?
        jnz     until_without_repeat_error   ; ...no, report error
        call    #eval_expr                   ; ...yes, evaluate control expression
        bis.w   r14, r15
        cmp.w   #0, r15                      ; result is true (non-zero)?
        jnz     x_until_10                   ; ...yes, exit the repeat loop
        mov.w   2(sp), lp                    ; ...no, restore interpreter context
        mov.w   4(sp), tp
        mov.b   -1(tp), r4                   ; prime R4 with character following REPEAT
        jmp     term                         ; ensure that statement is correctly terminated
x_until_10:
        add.w   #6, sp                       ; remove REPEAT context
        jmp     term                         ; ensure that statement is correctly terminated

; EXIT FOR
x_exit_for:
        call    #end_of_statement            ; anything after EXIT FOR?
        jnz     syntax_error                 ; ...yes, syntax error
        mov.w   #T_FOR, r11                  ; ...no, find a matching FOR loop
        jmp     x_exit_10

; EXIT REPEAT
x_exit_repeat:
        call    #end_of_statement            ; anything after EXIT REPEAT?
        jnz     syntax_error                 ; ...yes, syntax error
        mov.w   #T_REPEAT, r11               ; ...no, find a matching REPEAT loop
        jmp     x_exit_10

; EXIT WHILE
x_exit_while:
        call    #end_of_statement            ; anything after EXIT WHILE?
        jnz     syntax_error                 ; ...yes, syntax error
        mov.w   #T_WHILE, r11                ; ...no, find a matching WHILE loop

x_exit_10:
        cmp.w   @sp, r11                     ; matched structure?
        jz      x_exit_30                    ; ...yes, transfer control out of loop
        cmp.w   #T_END, @sp                  ; searched whole stack?
        jz      exit_not_in_loop_error       ; ...yes, not found so emit invalid exit
        cmp.w   #T_GOSUB, @sp
        jz      exit_not_in_loop_error
        cmp.w   #T_RESUME, @sp
        jz      exit_not_in_loop_error
        add.w   #6, sp                       ; 6 bytes are stacked for WHILE/REPEAT loops
        cmp.w   #T_FOR, @sp
        jnz     x_exit_10
        add.w   #6, sp                       ; 12 bytes are stacked for a FOR loop
        jmp     x_exit_10

x_exit_30:
        cmp.w   #T_WHILE, r11                ; WHILE-WEND loop?
        jnz     x_exit_40
        mov.w   #T_WEND, r11                 ; search for a matching WEND
x_exit_40:
        cmp.w   #T_FOR, r11                  ; FOR-NEXT loop?
        jnz     x_exit_50
        mov.w   #T_NEXT, r11                 ; search for a matching NEXT
x_exit_50:
        cmp.w   #T_REPEAT, r11               ; REPEAT-UNTIL loop?
        jnz     x_exit_60
        mov.w   #T_UNTIL, r11                ; search for a matching UNTIL
x_exit_60:
        mov.w   2(sp), lp                    ; get token pointer and line pointer of
        mov.w   4(sp), tp                    ; token immediately following WHILE/REPEAT/FOR
        call    #find_control_match
        jnc     exit_not_in_loop_error
        mov.b   @tp+, r4
        add.w   #6, sp                       ; remove structure parameters from stack
        cmp.w   #T_NEXT, r11                 ; FOR-NEXT loop stores a bit more context
        jnz     term
        add.w   #6, sp
        jmp     term

find_control_match:
        push.w  lp                           ; save line pointer for error message if we can't find the endpoint
        push.w  #T_END                       ; push delimiter
find_control_match_00:
        cmp.w   #T_WHILE, r4                 ; start of a WHILE-WEND?
        jz      open_control
        cmp.w   #T_FOR, r4                   ; start of a FOR-NEXT?
        jz      open_control
        cmp.w   #T_REPEAT, r4                ; start of a REPEAT-UNTIL?
        jz      open_control
        cmp.w   #T_END, @sp                  ; everything closed?
        jnz     find_control_match_10
        cmp.w   r4, r11                      ; matched the terminator we're looking for?
        jnz     find_control_match_10        ; no, try other terminators
        add.w   #4, sp                       ; remove T_END and lp from stack
        setc
        ret                                  ; return with LP and TP set appropriately and C=1

find_control_match_10:
        cmp.w   #T_NEXT, r4                  ; close FOR-NEXT structure?
        jz      close_for                    ; yes
        cmp.w   #T_WEND, r4                  ; close WHILE-WEND structure?
        jz      close_while                  ; yes
        cmp.w   #T_UNTIL, r4                 ; close REPEAT-UNTIL structure?
        jz      close_repeat                 ; yes
        jmp     open_control_00              ; no, consider next token

open_control:
        push.w  r4
open_control_00:
        cmp.b   #0, r4
        mov.b   @tp+, r4
        jnz     find_control_match_00
        cmp.w   #0, lp
        jz      unterminated_structure
        mov.w   @lp, lp
        mov.w   lp, tp
        add.w   #4, tp
        mov.b   @tp+, r4
        jmp     find_control_match_00
close_for:
        pop.w   r15
        cmp.w   #T_FOR, r15
        jnz     next_without_for_error
        jmp     open_control_00
close_while:
        pop.w   r15
        cmp.w   #T_WHILE, r15
        jnz     end_while_without_while_error
        jmp     open_control_00
close_repeat:
        pop.w   r15
        cmp.w   #T_REPEAT, r15
        jnz     until_without_repeat_error
        jmp     open_control_00

; Clear unclosed structures from stack
unterminated_structure:
        mov.w   @sp, r4                      ; get the structure which is unterminated
unterminated_structure_10:
        pop.w   r15
        cmp.w   #T_END, r15
        jnz     unterminated_structure_10
        pop.w   lp
        clrc
        ret

; END
x_end:
        call    #end_of_statement            ; anything after EXIT REPEAT?
        jnz     syntax_error                 ; ...yes, syntax error
        mov.w   #END_OF_RAM, sp              ; FIXME
        jmp     ready                        ; ended, so terminate

; FOR var = init TO final
x_for:
        call    #is_letter
        jnc     syntax_error
        mov.b   @tp+, r7
        cmp.w   #T_EQ, r7
        jnz     syntax_error
        sub.w   #'A', r4
        add.w   r4, r4
        add.w   r4, r4
        mov.w   r4, r11
        call    #get_eval_expr
        mov.w   r14, vars(r11)
        mov.w   r15, vars+2(r11)
        cmp.w   #T_TO, r4
        jnz     syntax_error
        call    #get_eval_expr
        push.w  r15                          ; limit           +10
        push.w  r14                          ;                 +8
        push.w  r11                          ; variable offset +6
        push.w  tp                           ; token pointer   +4
        push.w  lp                           ;                 +2
        push.w  #T_FOR
        jmp     term

x_next:
        cmp.w   #T_FOR, @sp                  ; is control structure FOR?
        jnz     next_without_for_error       ; ...no, report error
        mov.w   6(sp), r11                   ; pointer to variable
        add.w   #1, vars(r11)                ; increment variable
        addc.w  #0, vars+2(r11)
        mov.w   8(sp), r14                   ; get limit
        mov.w   10(sp), r15
        sub.w   vars(r11), r14               ; subtract current value
        subc.w  vars+2(r11), r15
        jnc     x_next_terminate_loop        ; if passed limit, terminate loop
        mov.w   4(sp), tp                    ; restore context
        mov.w   2(sp), lp
        mov.b   -1(tp), r4
        jmp     term                         ; continue with loop
x_next_terminate_loop:
        add.w   #6*2, sp                     ; remove FOR context
        jmp     term                         ; continue with statement after NEXT

; ON BREAK stmt
; ON X GOTO l1, l2, ... ELSE ...
x_on:
        cmp.w   #T_BREAK, r4                 ; check for ON BREAK
        jz      x_on_break
        call    #eval_expr                   ; evaluate control expression X
        mov.w   r14, r11
        cmp.w   #T_GOTO, r4                  ; require GOTO
        jnz     syntax_error
x_on_goto_00:
        mov.b   @tp+, r4                     ; get token after GOTO
        cmp.b   #0, r4                       ; on end of statement, continue execution
        jz      term
        cmp.b   #':', r4
        jz      term
        cmp.b   #T_ELSE, r4                  ; got to end of list without a match, but have an ELSE?
        jz      x_if_10                      ; ...yes, treat ELSE like IF's THEN allowing e.g. ELSE 100 or ELSE PRINT Y
        call    #eval_expr                   ; evaluate line number
        sub.w   #1, r11                      ; this label index matches control expression?
        jz      x_goto_00                    ; yes
        cmp.b   #',', r4                     ; more in the list
        jz      x_on_goto_00                 ; yep, try to match another
        cmp.b   #T_ELSE, r4                  ; got to end of list without a match, but have an ELSE?
        jz      x_if_10                      ; ...yes, treat ELSE like IF's THEN allowing e.g. ELSE 100 or ELSE PRINT Y
	jmp     term
x_on_break:
        mov.w   tp, &break_tp
        mov.w   lp, &break_lp
        jmp     x_rem

; IF expr THEN [stmt | number]
; IF expr [THEN] stmt
x_if:
        call    #eval_expr                   ; evaluate control expr
        bis.w   r14, r15                     ; controlling expression is false?
        cmp.w   #0, r15
        jz      x_rem                        ; ...yes, skip whole line
        cmp.w   #T_THEN, r4                  ; THEN?
        jnz     chrgot                       ; ...no, THEN is omitted, so next is a statement
x_if_10:
        mov.b   @tp+, r4                     ; get token following THEN
        call    #is_digit                    ; is it a digit?
        jc      x_goto                       ; ...yes, assume it's a line number and use GOTO
        jmp     chrgot

; INPUT [#channel, ] [string,] var [, string], var...
x_input:
        mov.w   #1, r15                      ; default to channel 1 if no channel number given
        call    #select_channel
        cmp.w   #'#', r4                     ; channel number being specified?
        jnz     x_input_10                   ; ...no, so just use default
        call    #get_eval_expr               ; evaluate the channel number
        mov.w   r14, r15                     ; only use the low 16 bits of the channel number
        call    #select_channel              ; select the channel, error if it's bad
x_input_00:
	call	#check_comma                 ; a comma must come next
x_input_10:
        cmp.b   #'"', r4                     ; string?
	jnz     x_input_20                   ; ...no, it must be a variable
	call    #print_quoted_string         ; print the prompt string
	call	#check_comma                 ; a comma must come next
	jmp     x_input_30                   ; do the input
x_input_20:
        mov.w   #'?', r15                    ; default prompt is just '?'
	call    #print_char
x_input_30:
	call	#varptr                      ; get variable we're inputting
	cmp.w   #4, r15                      ; only allow READ on word variables
	jnz     syntax_error
x_input_35:
	mov.w   r14, r11                     ; save for later use
	call    #get_string                  ; read the input to ramtop
	mov.w   &ramtop, r10                 ; get ready to parse input
x_input_40:
	mov.b   @r10+, r4                    ; get character from input buffer
	cmp.b   #' ', r4                     ; space?
	jz      x_input_40                   ; ...yes, skip them
	cmp.b   #0, r4                       ; nothing on line?
	jz      x_input_50                   ; ...no, try again
	call    #is_digit                    ; leading digit?
	jnc     x_input_50                   ; ...no, print bad data error and try again
        push.w  tp                           ; save interpreter's token pointer
	mov.w   r10, tp                      ; prepare to parse number from buffer
	call    #read_int                    ; do the parsing
	mov.w   r14, @r11                    ; store read value
	mov.w   r15, 2(r11)
	pop.w   tp                           ; restore interpreter's token pointer
	mov.b   -1(tp), r4                   ; refresh context
	call    #end_of_statement            ; end of statement looming?
	jz      term                         ; ...yes, execute next statement
	jmp     x_input_00
x_input_50:
	mov.w   #bad_input_error_message, r15; print bad input error
	call    #print_message
        mov.w   #'?', r15                    ; repeat default prompt
	call    #print_char
	jmp     x_input_35                   ; try again

; PRINT expr [, | ;] expr...
x_print:
        mov.w   #0, r11                      ; r11 is "hold CR" flag
        mov.w   #1, r15                      ; default to channel 1 if no channel number given
        call    #select_channel
        cmp.w   #'#', r4                     ; channel number being specified?
        jnz     x_print_02                   ; ...no, so just use default
        call    #get_eval_expr               ; evaluate the channel number
        mov.w   r14, r15                     ; only use the low 16 bits of the channel number
        call    #select_channel              ; select the channel, error if it's bad
        jmp     x_print_50                   ; see if ',' or anything terminates the statement early
x_print_00:
	mov.b	#9, r15                      ; print tab on ','
	call	#print_char
x_print_01:
        mov.b   @tp+, r4                     ; get next token
x_print_02:
        cmp.w   #T_AT, r4                    ; AT?
        jnz     x_print_03                   ; ...no, continue to decode it
        call    #get_eval_expr               ; evaluate positional expression
        call    #print_at                    ; position cursor
        jmp     x_print_40                   ; deal with terminators
x_print_03:
        cmp.b   #'"', r4                     ; printing a string?
        jne     x_print_20                   ; ...no, continue to decode it
	call    #print_quoted_string         ; print the string
        jmp     x_print_35                   ; continue parsing
x_print_10:
        mov.b   @tp+, r4                     ; get next token
x_print_20:
        cmp.b   #T_HEX, r4                   ; print value in hexadecimal?
        jnz     x_print_23                   ; ...no, continue to decode it
        call    #get_eval_expr               ; evaluate expression to print
        call    #print_hex                   ; print it in hex
        jmp     x_print_35                   ; deal with termimators
x_print_23:
        cmp.b   #T_BIN, r4                   ; print value in binary?
        jnz     x_print_25                   ; ...no, continue to decode it
        call    #get_eval_expr               ; evaluate expression to print
        call    #print_bin                   ; print it in binary
        jmp     x_print_35                   ; deal with termimators
x_print_25:
        cmp.b   #T_CLEAR, r4                 ; CLEAR?
        jnz     x_print_30                   ; ...no, continue to decode it
        mov.w   #0x0c, r15                   ; prepare to print form feed
        call    #print_char                  ; send it to the channel
        jmp     x_print_40                   ; deal with separators
x_print_30:
	call    #end_of_statement            ; end of line which terminates printing?
        jz      x_print_end                  ; ...yes, wrap up
        call    #eval_expr                   ; nothing matches, so parse it as an expression
        call    #print_integer               ; print it in decimal
x_print_35:
        sub.w   #1, tp                       ; back up token pointer for code ahead
x_print_40:
        mov.w   #1, r11                      ; prepare to hold CRLF
        mov.b   @tp+, r4                     ; prime R4 with next character
x_print_50:
        cmp.b   #';', r4                     ; expression separator?
        jz      x_print_01                   ; ...yes, continue with hold flag set
        cmp.b   #',', r4                     ; expression separator?
        jz      x_print_00                   ; ...yes, continue with hold flag set
        cmp.w   #0, r4                       ; end of line?
        jz      x_print_end_10               ; yes, terminate with a new line
        cmp.w   #':', r4                     ; end if statement?
        jz      x_print_end_10               ; yes, terminate with a new line
        jmp     syntax_error                 ; bad termination is reported as a syntax error

x_print_end:
        cmp.w   #0, r11                      ; hold CRLF?
        jnz     term                         ; ...yes, forget new line and continue with execution
x_print_end_10:
        mov.w   #'\n', r15                   ; print a new line
        call    #print_char
        jmp     term                         ; continue with processing

; Print the string delimited by a quotation mark.
; Shared between INPUT and PRINT.
print_quoted_string:
        mov.b   @tp+, r4                     ; get character in string
        cmp.b   #'"', r4                     ; terminated?
        jz      print_quoted_string_return   ; ...yes, deal with terminators
        cmp.b   #0, r4                       ; end of string?
        jz      mismatched_quote_error       ; ...yes, report error
        mov.w   r4, r15                      ; prepare character to print
        call    #print_char                  ; print it
        jmp     print_quoted_string          ; and get the next one
print_quoted_string_return:
        mov.b   @tp+, r4                     ; fetch token after final quotation mark
	ret

; BPUT [#channel, ] expr, expr, ...
x_bput:
        mov.w   #1, r15                      ; default to channel 1 if no channel number given
        call    #select_channel
        cmp.w   #'#', r4                     ; channel number being specified?
        jnz     x_bput_10                    ; ...no, so just use default
        call    #get_eval_expr               ; evaluate the channel number
        mov.w   r14, r15                     ; only use the low 16 bits of the channel number
        call    #select_channel              ; select the channel, error if it's bad
x_bput_00:
	call	#check_comma                 ; a comma must come next
x_bput_10:
	call	#eval_expr                   ; get the byte to put to the channel
	mov.w   r14, r15                     ; the byte is delivered in R14, so move to R15
	call	#print_char                  ; put the byte
	call	#end_of_statement            ; finished statement?
	jnz	x_bput_00                    ; ...no, need a comma and another expression
	jmp	term                         ; finish off statement

; BGET #channel
x_bget:
	cmp.b	#'#', r4                     ; require a channel number
	jnz	syntax_error                 ; ...but didn't get one
	mov.b   @tp+, r4                     ; skip '#'
	call    #eval_unary                  ; get the channel number
        mov.w   r14, r15                     ; only use the low 16 bits of the channel number
	call    #select_channel              ; select the channel
	call	#get_char                    ; read the character into R15
	mov.w	r15, r14                     ; zero-extend TOS
	mov.w   #0, r15
	ret

; RESET
x_reset:
        mov.w   #0, &WDTCTL                  ; bad password generates reset
        jmp     term                         ; ...but simulator doesn't

; MEMORY
x_memory:
        call    #memory
        jmp     term

memory:
        mov.w   sp, r14
        sub.w   &ramtop, r14
        mov.w   #0, r15
        call    #print_integer
        mov.w   #ram_message, r15
        call    #print_message
        mov.w   #0xfc00, r14
        sub.w   &flashtop, r14
        call    #print_integer
        mov.w   #flash_message, r15
        call    #print_message
        ret

; BREAK [ ON | OFF | expr ]
x_break:
        call    #eval_expr_on_off            ; Evaluate ON to -1, OFF to 0, and
        bic.w   #BREAK, &flags               ; Prepare for BREAK OFF
        bis.w   r14, r15
        cmp.w   #0, r15                      ; zero value means BREAK OFF
        jz      x_break_return
        bis.w   #BREAK, &flags               ; non-zero value turns on BREAK flag
x_break_return:
        jmp     term

; RESUME
x_resume:
        pop.w   r4                           ; get opened structure
        cmp.w   #T_RESUME, r4                ; RESUME expected from ON ... processing?
        jnz     cant_resume_error            ; ...no, can't RESUME
        pop.w   lp                           ; restore resume point
        pop.w   tp
        sub.w   #1, tp
        pop.w   sr                           ; restore interrupt enable status
        jmp     chrget                       ; continue with mainline code

; INTERRUPT ON
; INTERRUPT OFF
x_interrupt:
        cmp.b   #T_ON, r4
        jnz     x_interrupt_10
        eint
        mov.b   @tp+, r4
        jmp     term
x_interrupt_10:
        cmp.b   #T_OFF, r4
        jnz     syntax_error
        dint
        mov.b   @tp+, r4
        jmp     term

; Evaluate an expression with ON evaluating to -1 and OFF evaluating to 0
eval_expr_on_off:
        cmp.w   #T_ON, r4                    ; ON?
        jz      eval_expr_on_off_10          ; ...yes, return TRUE
        cmp.w   #T_OFF, r4                   ; OFF?
        jnz     eval_expr                    ; ...no, parse as an expression
        mov.b   @tp+, r4                     ; skip OFF token
        jmp     x_false                      ; return FALSE
eval_expr_on_off_10:
        mov.b   @tp+, r4                     ; skip ON token
        jmp     x_true                       ; return TRUE

; Get next token and evaluate expression
get_eval_expr:
        mov.b   @tp+, r4                     ; get next token

; Evaluate expression and return an integer in R15:R14
eval_expr:
        push.w  #T_PLUS-1                    ; least binding operator, lower precedence than any other

eval_expr_10:
	call    #eval_unary                  ; evaluate a unary expression

; Operator?
eval_expr_cont:
        cmp.w   #FIRST_BINARY_TOKEN, r4      ; if this isn't a binary operator, reduce whole stack and wrap up
        jnc     eval_expr_60
        cmp.w   #LAST_BINARY_TOKEN+1, r4
        jc      eval_expr_60

        mov.w   @sp, r13                     ; get stacked operator and compare precendence of this operator
        cmp.b   operator_precedence-FIRST_BINARY_TOKEN+1(r4), operator_precedence-FIRST_BINARY_TOKEN+1(r13)
        jc      eval_expr_60                 ; same or lower precedence, so reduce and keep parsing
        push.w  r15                          ; save TOS value on stack
        push.w  r14
        push.w  r4                           ; save current operator on stack
        mov.b   @tp+, r4                     ; get next character from the token stream
        jmp     eval_expr_10                 ; and it must start a unary expression

eval_expr_60:
        pop.w   r10                          ; get stacked operator
        cmp.w   #FIRST_BINARY_TOKEN-1, r10   ; if it's the lowest precedence (cf eval_expr), we're done.
        jz      eval_expr_return
        pop.w   r12                          ; pop saved TOS into NOS
        pop.w   r13
        add.w   r10, r10                     ; compute index into operator table
	                                     ; invoke operator routine
        call    binarytab-FIRST_BINARY_TOKEN*2(r10) 
        jmp     eval_expr_cont               ; continue with parsing expression
eval_expr_return:
        ret

get_eval_unary:
	mov.b   @tp+, r4                     ; get next token

eval_unary:
        tst.b   r4
        jge     eval_expr_20

; Token is negative, so it's an operator or function
        cmp.w   #T_BIT, r4                   ; BIT used as unary operator?
	jz      x_unary_bit                  ; ...yes, handle it specially
        cmp.w   #T_MINUS, r4                 ; - used as unary operator?
	jz      x_negate                     ; ...yes, handle it specially
        cmp.w   #T_PLUS, r4                  ; + used as unary operator?
	jz      get_eval_unary               ; ...yes, handle it specially

        cmp.w   #FIRST_UNARY_TOKEN, r4
        jnc     syntax_error
        cmp.w   #LAST_UNARY_TOKEN+1, r4
        jc      syntax_error
        add.w   r4, r4
        mov.w   r4, r13
        mov.b   @tp+, r4
        br      unarytab-FIRST_UNARY_TOKEN*2(r13)

eval_expr_20:
        cmp.b   #'(', r4
        jne     eval_expr_30
        call    #get_eval_expr
        cmp.b   #')', r4
        jne     mismatched_parenthesis_error
        mov.b   @tp+, r4
	ret

eval_expr_30:
        call    #is_digit
        jc      read_int

eval_expr_50:
	call    #varptr                      ; parse variable and get address into R14
	cmp.w   #4, r15                      ; address of a 32-bit value?
	jz      eval_expr_70                 ; ...yes, read 32-bit value
	cmp.w   #2, r15                      ; address of a 16-bit value?
	jz      eval_expr_80                 ; ...yes, read 16-bit value
	mov.b   @r14, r14                    ; ...no, read byte
	mov.w   #0, r15                      ; zero extend
	ret
eval_expr_70:
	mov.w   2(r14), r15                  ; fetch 32 bits
	mov.w   @r14, r14
	ret
eval_expr_80:
	mov.w   @r14, r14                    ; fetch 16 bits
	mov.w   #0, r15                      ; zero extend
	ret

; Read integer value into R15:R14 from the token stream.
read_int:
        mov.w   #0, r15
        mov.w   #0, r14
read_int_10:
        sub.w   #'0', r4
        call    #mul_10
        add.w   r4, r14
        addc.w  #0, r15
        mov.b   @tp+, r4
        call    #is_digit
        jc      read_int_10
        ret

; >
x_gt:
        sub.w   r12, r14
        subc.w  r13, r15
        jl      x_true
        jmp     x_false

; <
x_lt:
        sub.w   r14, r12                     ; compare operands, less?
        subc.w  r15, r13
        jl      x_true                       ; ...yes, return true else fall through

; FALSE
x_false:
        mov.w   #0, r15                      ; FALSE is zero
        mov.w   #0, r14
        ret

; =
x_eq:
        cmp.w   r14, r12                     ; same low parts?
        jnz     x_false                      ; ...no, must be different
        cmp.w   r15, r13                     ; same high parts?
        jnz     x_false                      ; ...no, return false else fall through

; TRUE
x_true:
        mov.w   #-1, r15                     ; true value is all bits set
        mov.w   #-1, r14
        ret

; <>
x_ne:
        cmp.w   r14, r12                     ; same low parts?
        jnz     x_true                       ; ...no, must be true
        cmp.w   r15, r13                     ; same high parts?
        jnz     x_true                       ; ...no, must be true
        jmp     x_false                      ; both parts are the same, so result is false

; a <= b = !(b > a)
x_le:
        sub.w   r12, r14
        subc.w  r13, r15
        jl      x_false
        jmp     x_true

; a >= b = !(b < a)
x_ge:
        sub.w   r14, r12
        subc.w  r15, r13
        jl      x_false
        jmp     x_true

; x BIT y
; If bit is not in 0..31, return 0.
x_bit:
        cmp.w   #0, r15
        jnz     x_false
        cmp.w   #0x20, r14
        jc      x_false

; If bit is within the high word, shift and adjust bit number.
        cmp.w   #16, r14
        jnc     x_bit_10
        mov.w   r13, r12

; Test bit and return value
x_bit_10:
        and.w   #0xf, r14                    ; now testing bits 0 through 15
        add.w   r14, r14
        and.w   bitpos(r14), r12             ; mask off uninteresting bits
        jz      x_false                      ; bit is zero
        jmp     x_true                       ; bit is one

bitpos:
        dw      1<<0, 1<<1, 1<<2, 1<<3, 1<<4, 1<<5, 1<<6, 1<<7
        dw      1<<8, 1<<9, 1<<10, 1<<11, 1<<12, 1<<13, 1<<14, 1<<15

; Binary + operator
x_plus:
        add.w   r12, r14                     ; add NOS to TOS
        addc.w  r13, r15
        ret

; Binary - operator
x_minus:
        sub.w   r14, r12                     ; subtract NOS from TOS
        subc.w  r15, r13
        mov.w   r12, r14
        mov.w   r13, r15
        ret

; Binary * operator
x_star:
#ifdef   HARDWARE_MULTIPLY
        
; Save status register so we can restore IE on exit.
        push.w  sr

; Disable interrupts and really wait for them to be disabled.
        dint
        nop

; Form low product (low1 * low2).
        mov.w   r12, &MPY
        mov.w   r14, &OP2
        mov.w   r14, &MAC                    ; prepare for high1*low2 + carry-over
        mov.w   &RESLO, r14

; Carry over high part of 32-bit product to when we calculate
; the high order product.
        mov.w   &RESHI, &RESLO
        mov.w   r13, &OP2
        mov.w   r12, &MAC                    ; low1*high2 + high1*low2 + carry over
        mov.w   r15, &OP2
        mov.w   &RESLO, r15

; Restore SR and return.  Isn't the multiplier cute?
        reti

#else

; Multiplication by clockwork.  Save working registers.
        push.w  r10
        push.w  r11
         
; Clear product
        clr.w   r10
        clr.w   r11

; Single bit multiplication step
L$0     clrc
        rrc.w   r15
        rrc.w   r14
        jnc     L$1

; One bit shifted out, update product
        add.w   r12, r10
        addc.w  r13, r11

; Update shift register
L$1     add.w   r12, r12
        addc.w  r13, r13

; Continue until all bits shifted out of multiplicand.
        cmp.w   #0, r14
        jne     L$0
        cmp.w   #0, r15
        jne     L$0

; Move product to result registers.
        mov.w   r10, r14
        mov.w   r11, r15

; Restore working registers
        pop.w   r11
        pop.w   r10
        ret
#endif

; Binary / operator
x_slash:
; Exchange registers.
        xor.w   r12, r14
	xor.w   r14, r12
	xor.w   r12, r14
        xor.w   r13, r15
	xor.w   r15, r13
	xor.w   r13, r15

; Check for division by zero.
x_slash_10:
        cmp.w   #0, r13
        jnz     x_slash_20
        cmp.w   #0, r12
        jz      divide_by_zero_error

; Save working registers.
x_slash_20:
        push.w  r9
        push.w  r10
        push.w  r11

; Quotient is negative if the dividend and divisor signs differ.
        mov.w   r15, r9
        xor.w   r13, r9

; Compute magnitude of divisor.
        tst.w   r13
        jge     Positive$Divisor
        xor.w   #-1, r12
        xor.w   #-1, r13
        add.w   #1, r12
        addc.w  #0, r13

; Compute magnitude of dividend.
Positive$Divisor
        tst.w   r15
        jge     Positive$Dividend
        xor.w   #-1, r14
        xor.w   #-1, r15
        add.w   #1, r14
        addc.w  #0, r15

; Perform division, r15:r14 = quotient, r11:r10 = remainder
Positive$Dividend
        call    #___uint32_divmod

; If the dividend and divisor signs differ, negate quotient.
        tst.w   r9
        jge     Positive$Quotient
        xor.w   #-1, r14
        xor.w   #-1, r15
        add.w   #1, r14
        addc.w  #0, r15

; Restore working registers.
Positive$Quotient
        pop.w   r11
        pop.w   r10
        pop.w   r9
        ret

; Binary MOD operator
x_mod:
; Exchange arguments
        xor.w   r12, r14
	xor.w   r14, r12
	xor.w   r12, r14
        xor.w   r13, r15
	xor.w   r15, r13
	xor.w   r13, r15

; Check for division by zero.
        cmp.w   #0, r13
        jnz     x_mod_10
        cmp.w   #0, r12
        jz      divide_by_zero_error

; Save working registers.
x_mod_10:
        push.w  r11
        push.w  r10
        push.w  r9

; Sign of remainder takes sign of dividend.
        mov.w   r15, r9
        
; Compute magnitude of divisor.
        tst.w   r12
        jge     x_mod_20
        xor.w   #-1, r12
        xor.w   #-1, r13
        add.w   #1, r12
        addc.w  #0, r13

; Compute magnitude of dividend.
x_mod_20:
        tst.w   r15
        jge     x_mod_30
        xor.w   #-1, r14
        xor.w   #-1, r15
        add.w   #1, r14
        addc.w  #0, r15

; Perform division, r15:r14 = quotient, r11:r10 = remainder
x_mod_30:
        call    #___uint32_divmod

; Move remainder to result registers.
        mov.w   r10, r14
        mov.w   r11, r15

; Apply sign of dividend to remainder.
        tst.w   r9
        jge     x_mod_40
        call	#negate

; Restore working registers.
x_mod_40:
        pop.w   r9
        pop.w   r10
        pop.w   r11
        ret
        
___uint32_divmod:

; Save working registers
        push.w  r8
        push.w  r9

; Zero remainder.
        clr.w   r11
        clr.w   r10

; Prime division step count.
        mov.w   #32+1, r9
        jmp     NextStep$

; Trial subtract.
DivideStep$
        sub.w   r12, r10
        subc.w  r13, r11
        jc      NextStep$

; Trial subtraction underflowed, so restore.
        add.w   r12, r10
        addc.w  r13, r11
        clrc

; Carry contains whether subtraction was made or not (inverted).
NextStep$
        rlc.w   r14
        rlc.w   r15
        rlc.w   r10
        rlc.w   r11

; Save carry out as remainder overshoots by one bit in the last step.
        rrc.w   r8

; Iterate for 32 bits.
        dec.w   r9
        jnz     DivideStep$
        
; Restore bit shifted out of remainder.
        rlc.w   r8
        rrc.w   r11
        rrc.w   r10

; Restore working registers.
        pop.w   r9
        pop.w   r8

; Resturn with quotient in r12:r13 and remainder in r14:r15.
        ret

; AND operator
x_and:
        and.w   r12, r14
        and.w   r13, r15
        ret

; BIC operator
x_bic:
        bic.w   r14, r12
        bic.w   r15, r13
	mov.w   r12, r14
	mov.w   r13, r15
        ret

; IMP operator - a IMP b == NOT a OR b
x_imp:
        xor.w   #-1, r12
        xor.w   #-1, r13                     ; Fall through to OR

; OR operator
x_or:
        bis.w   r12, r14
        bis.w   r13, r15
        ret

; XOR operator
x_xor:
        xor.w   r12, r14
        xor.w   r13, r15
        ret

; EQV operator - a EQV b == NOT (a XOR b)
x_eqv:
        xor.w   r12, r14
        xor.w   r13, r15
	jmp     x_not_10

; NOT operator
x_not:
        call    #eval_unary                  ; evaluate argument
x_not_10:
        xor.w   #-1, r14
        xor.w   #-1, r15
        ret

; Unary - operator
x_negate:
        call    #get_eval_unary              ; evaluate argument
negate:
        xor.w   #-1, r14                     ; negate by forming one's complement
        xor.w   #-1, r15
        add.w   #1, r14                      ; and adding one to form two's complement
        addc.w  #0, r15
x_negate_return:
        ret

; VARPTR addr
x_varptr:
        call    #varptr                      ; get address
	mov.w   #0, r15                      ; overwrite size designation
	ret

; Parse a variable and return its address in R14 and zero in R15.
; The variable can be simple or an array
varptr:
        mov.w   #0, r14                      ; set the array base as zero for ?
	cmp.b   #'?', r4                     ; byte indirection operator?
        jz      varptr_25                    ; ...yes, do byte indirection
        cmp.b   #'@', r4                     ; word indirection operator?
        jnz     varptr_05
        mov.b   @tp+, r4                     ; skip '@'
        call    #eval_unary                  ; get address
        bic.w   #1, r14                      ; word align
	mov.w   #2, r15                      ; indicate 16-bit data
	ret
varptr_05:
        cmp.b   #'!', r4                     ; word indirection operator?
        jnz     varptr_07                    ; ...no, must start with an variable or array name
        mov.b   @tp+, r4                     ; skip '!'
        call    #eval_unary                  ; get address
        bic.w   #1, r14                      ; word align
	mov.w   #4, r15                      ; indicate 32-bit data
	ret
varptr_07:
        call    #is_letter                   ; variable or index name?
        jnc     syntax_error                 ; ...no, syntax error
        sub.w   #'A', r4                     ; compute index into variable or array table
        add.w   r4, r4
        add.w   r4, r4
	cmp.b   #'(', @tp                    ; array reference?
	jz      varptr_10                    ; ...yes, handle array assignment
	add.w   #vars, r4                    ; form address of variable
	mov.w   r4, r14                      ; return in R15
	mov.b   @tp+, r4                     ; get token following variable
	jmp     varptr_20                    ; consider suffix operators
varptr_10:
	cmp.w   #0, arrays(r4)               ; has array been dimensioned with DIM?
	jz      array_not_dimensioned_error  ; ...no, report not dimensioned error
	push.w  r4                           ; save array table index
	mov.b   @tp+, r4                     ; get '('
	call    #eval_unary                  ; evaluate index
	cmp.w   #0, r15                      ; index negative or just too big?
	jnz     index_out_of_bounds_error    ; ...yes, out of bounds error
	pop.w   r15                          ; restore array table index
	cmp.w   arrays+2(r15), r14           ; index less than dimensioned size?
	jc      index_out_of_bounds_error    ; ...no, out of bounds error
	add.w   r14, r14                     ; compute index into array
	add.w   r14, r14
	add.w   arrays(r15), r14             ; compute address of element in the array
varptr_20:
	cmp.b   #'?', r4                     ; byte index operator?
	jnz     varptr_30                    ; ...no, consider next suffix
	mov.w   @r14, r14                    ; fetch indexed value which is the array base
varptr_25:
	mov.b   @tp+, r4                     ; skip '?'
	push.w  r14                          ; save base address
	call    #eval_unary                  ; get index
	pop.w   r15                          ; restore base address
	add.w   r15, r14                     ; calculate indexed address
	mov.w   #1, r15                      ; indicate byte address
	ret
varptr_30:
	cmp.b   #'!', r4                     ; word index operator?
	jnz     varptr_40                    ; ...no, consider next suffix
	mov.b   @tp+, r4                     ; skip '?'
	mov.w   @r14, r14                    ; fetch indexed value which is the array base
	push.w  r14                          ; save base address
	call    #eval_unary                  ; get index
	add.w   r14, r14                     ; calculate word index
	add.w   r14, r14
	pop.w   r15                          ; restore base address
	add.w   r15, r14                     ; calculate indexed address
	jmp     varptr_20                    ; see if more suffixes
varptr_40:
        mov.w   #4, r15                      ; indicate word address
	ret

; ABS operator
x_abs:
        call    #eval_unary                  ; evaluate argument
abs:
        cmp.w   #0, r15                      ; positive or zero?
        jge     x_negate_return              ; ...yes, nothing to be done
	jmp     negate                       ; ...no, negate top of stack

; Unary BIT operator
x_unary_bit:
        mov.b   @tp+, r4                     ; enter with r4=BIT, which differs from regular routines
        call    #eval_unary                  ; get bit number
        and.w   #31, r14                     ; take modulo 32
        add.w   r14, r14                     ; compute index into bit table
        cmp.w   #32, r14                     ; bit 16 to 32 required?
        jc      x_unary_bit_10               ; ...yes, load high part and zero low part
        mov.w   bitpos(r14), r14             ; load bit pattern to low half of word
        mov.w   #0, r15                      ; zero high half
        ret
x_unary_bit_10:
        mov.w   bitpos-32(r14), r15          ; load high part of bit pattern
        mov.w   #0, r14                      ; zero low half
	ret

; SGN operator
x_sgn:
        call    #eval_unary                  ; evaluate argument
        cmp.w   #0, r15                      ; negative?
        jn      x_true                       ; ...yes, return -1
        bis.w   r14, r15                     ; know argument > 0, combine argument halves
        cmp.w   #0, r15                      ; argument is exactly zero?
        jz      x_sgn_return                 ; ...yes, nothing to do as SGN(0) = 0
        mov.w   #0, r15                      ; argument > 0, so return 1
        mov.w   #1, r14
x_sgn_return:
        ret

; RANDOMIZE expr
x_randomize:
	call    #eval_expr                   ; get user's seed value
	mov.w   r14, &rndseed[0]             ; store new seed
	mov.w   r15, &rndseed[1]
	jmp     term                         ; check statement terminated correctly

; RND function, based on 32-bit shift register with feedback
; from bits 31 and 26.
x_rnd:
        cmp.w   #'(', r4                     ; RND with an argument?
	jz      x_rnd_10                     ; ...yes, take a random number modulo that argument
rnd:
	mov.w   &rndseed[0], r14             ; get current seed
	mov.w   &rndseed[1], r15
	mov.w	#32, r12                     ; generate another 32 bits
rnd_10:
	mov.w   r15, r13                     ; move to working register
	and.w   #0x4800, r13                 ; form exclusive or of bits 31 and 26 of LFSR
	add.w   #0x3800, r13                 ; you'll need to hand-crank this code to see
	bit.w   #0x4000, r13                  ; the tricks in action...
	addc.w  r14, r14                     ; shift through LFSR
	addc.w  r15, r15
	sub.w   #1, r12                      ; got 32 new pseudo-random bits?
	jnz     rnd_10                       ; ...no, crank out another one
	mov.w   r14, &rndseed[0]             ; store new seed
	mov.w   r15, &rndseed[1]
	ret
x_rnd_10:
	call    #eval_unary                  ; evaluate the modulus
	push.w  r15                          ; save it on the stack
	push.w  r14
	call	#rnd                         ; get a random number
	and.w   #0x7fff, r15                 ; ensure it's positive
	pop.w   r12                          ; restore value to NOS
	pop.w   r13
	jmp     x_mod_10                     ; reduce random number to requested range

; PEEK operator
x_peek:
        call    #eval_unary                  ; evaluate argument
        mov.b   @r14, r14                    ; peek byte
        mov.w   #0, r15                      ; zero extend
        ret

; DPEEK operator
x_dpeek:
        call    #eval_unary                  ; evaluate argument
        bic.w   #1, r14                      ; word align pointer
        mov.w   @r14, r14                    ; peek work
        mov.w   #0, r15                      ; zero extend
        ret

; TOP function
x_top:
        mov.w   &ramtop, r14                 ; get top of used RAM
        mov.w   #0, r15                      ; zero extend
        ret

; GOTO line
x_goto:
        call    #eval_expr
x_goto_00:                                   ; Entry point for ON ... GOTO
        call    #find_line
        mov.w   r15, lp
        mov.w   r15, tp
        jmp     next_line

; GOSUB line
x_gosub:
        call    #eval_expr                   ; evaluate line number
        call    #find_line                   ; find the target line, or error if not found
        push.w  lp                           ; stack interpreter state
        push.w  tp
        push.w  #T_GOSUB                     ; push context marker as a GOSUB needing a RETURN
        mov.w   r15, lp                      ; go to the target line
        mov.w   r15, tp
        jmp     next_line

; RETURN
x_return:
        pop.w   r11                          ; get top context marker
        cmp.w   #T_GOSUB, r11                ; is top context a GOSUB?
        jne     return_without_gosub_error   ; ...no, RETURN without matching GOSUB
        pop.w   tp                           ; restore state
        pop.w   lp
        jmp     term                         ; ensure statement terminated correctly

; TRACE <expr>
; TRACE #<expr>
x_trace:
        cmp.b   #'#', r4
        jz      x_trace_10
        call    #eval_expr
        and.w   #1, r14
        mov.b   r14, trace_flags
        jmp     term
x_trace_10:
        call    #get_eval_expr
        mov.w   r14, r15
        call    #validate_channel
        mov.b   r15, &trace_channel
        jmp     term

; LIST
x_list:
        mov.w   #0, r8                       ; Set indent to zero.
        mov.w   #1, r15                      ; Select channel #1 by default.
        call    #select_channel

; Parse channel specification for listing redirection.
        cmp.b   #'#', r4
        jnz     x_list_00
        call    #get_eval_expr
        mov.w   r14, r15
        call    #select_channel

; Start listing from the start.
x_list_00:
        mov.w   &program_start, r11
        cmp.w   #0, r11
        jz      term                         ; If no program, finished listing
x_list_03:
        mov.w   2(r11), r14
        mov.w   #0, r15
        call    #print_integer
        mov.b   4(r11), r15                  ; get first token on line
        push.w  r8
        call    #list_outdent                ; outdent if it closes a loop
        jmp     x_list_03a
x_list_03b:
        mov.b   #' ', r15
        call    #print_char
x_list_03a:
        sub.w   #1, r8
        jc      x_list_03b
        pop.w   r8
        mov.b   #0, r15
        mov.w   r11, r12
        add.w   #4, r12
x_list_04:
        call    #list_space
x_list_05:
        mov.b   @r12+, r15
        cmp.b   #0, r15
        jz      x_list_20
        jge     x_list_10

        call    #list_outdent

; Indent on FOR, WHILE, REPEAT
        cmp.w   #T_FOR, r15
        jz      x_list_indent
        cmp.w   #T_WHILE, r15
        jz      x_list_indent
        cmp.w   #T_REPEAT, r15
        jnz     x_list_05a
x_list_indent:
        add.w   #2, r8
x_list_05a:
        call    #list_space
x_list_06:
        mov.w   #kwtab, r13
        and.w   #0x7f, r15
        jmp     x_list_08
x_list_07:
        mov.b   @r13+, r14
        cmp.b   #0, r14
        jnz     x_list_07
x_list_08:
        sub.w   #1, r15
        jc      x_list_07
x_list_09:
        mov.b   @r13+, r15
        cmp.b   #0, r15
        jz      x_list_04
        call    #list_print
        jmp     x_list_09
x_list_10:
        call    #list_print
        cmp.b   #',', r15
        jz      x_list_04
        jmp     x_list_05
x_list_20:
        mov.b   #'\n', r15
        call    #list_print
        mov.w   @r11, r11
        cmp.w   #0, r11
        jne     x_list_03
        jmp     term

; Output r15 and save it in r9 as the last character output.
list_print:
        mov.w   r15, r9
        jmp     print_char

; Output a space if the last wasn't a space.
list_space:
        cmp.w   #' ', r9                     ; was the last character output a space
        jz      list_space_return            ; yes, don't output another space
        push.w  r15
        mov.w   #' ', r15
        call    #print_char
        mov.w   #' ', r9
        pop.w   r15
list_space_return:
        ret

; Outdent if token is NEXT, WEND, or UNTIL:
list_outdent:
        cmp.w   #T_NEXT, r15
        jz      list_outdent_10
        cmp.w   #T_WEND, r15
        jz      list_outdent_10
        cmp.w   #T_UNTIL, r15
        jnz     list_outdent_return
list_outdent_10:
        cmp.w   #2, r8
        jnc     list_outdent_return
        sub.w   #2, r8
list_outdent_return:
        ret

; AUTO
x_auto:
	mov.w	#10, &auto_line              ; start at 10
	mov.w   #10, &auto_inc               ; and increment in 10s
	call	#end_of_statement            ; end of statement, i.e. default?
	jz	term                         ; ...yes, return to interpreter
	call	#eval_expr                   ; read starting line number
	call    #abs                         ; apply sanity filter
	mov.w   r14, &auto_line              ; save it
	cmp.b   #',', r4                     ; increment present?
	jnz     term                         ; ...no, end of statement
	call    #get_eval_expr               ; get increment
	call    #abs                         ; don't go backwards
	cmp.w   #0, r14                      ; stupid increment of zero?
	jz      term                         ; ...yes, keep default of 10
	mov.w   r14, &auto_inc               ; save it
	jmp     term

; STOP
x_stop:
        cmp.w   #0, lp                       ; STOP in immediate mode says can't continue
        jnz     x_stop_10
        mov.w   #1, tp
x_stop_10:
        sub.w   #1, tp
        mov.w   tp, &cont_tp
        mov.w   lp, &cont_lp
        mov.b   #1, &current_channel
        mov.w   #break_in_message, r15
        call    #print_error_message
        jmp     ready

; CONTINUE
x_continue:
        cmp.w   #0, &cont_tp                 ; anywhere to continue from?
        jz      cant_continue_error          ; ...no, can't continue error
        mov.w   &cont_tp, tp                 ; restore continutation context to interpreter context
        mov.w   &cont_lp, lp
        mov.w   #0, &cont_tp                 ; obliterate continue context
        mov.w   #0, &cont_lp
        jmp     chrget                       ; note special entry point, not "term" as context has changed

; Multiply R15:R14 by 10, corrupts R13:R12
mul_10:
        mov.w   r14, r12
        mov.w   r15, r13
        add.w   r14, r14  ; *2
        addc.w  r15, r15
        add.w   r14, r14
        addc.w  r15, r15  ; *4
        add.w   r12, r14
        addc.w  r13, r15  ; *5
        add.w   r14, r14
        addc.w  r15, r15  ; *10
        ret

; Ensure the current token is a comma and consume it
check_comma:
        cmp.b   #',', r4
        jnz     syntax_error
        mov.b   @tp+, r4
        ret

; Returns with C=1 if R4 is a digit
is_digit:
        cmp.w   #'0', r4
        jnc     is_digit_00
        cmp.w   #'9'+1, r4
        xor.w   #1, sr
is_digit_00
        ret

; Returns with C=1 if R4 is a letter
is_letter:
        cmp.w   #'A', r4
        jnc     is_letter_00
        cmp.w   #'Z'+1, r4
        xor.w   #1, sr
is_letter_00:
        ret

; Change R4 to uppercase if it's lowercase
uppercase:
        cmp.w   #'a', r4
        jnc     uppercase_00
        cmp.w   #'z', r4
        bic.w   #0x20, r4
uppercase_00:
        ret

; Set R15 to the line number in R14, or zero if no such line.
find_line:
        mov.w   &program_start, r15
        jmp     find_line_10
find_line_00:
        cmp.w   #0, r15
        jz      find_line_20
        mov.w   @r15, r15
find_line_10:
        cmp.w   2(r15), r14
        jz      uppercase_00
        jc      find_line_00
find_line_20:
        mov.w   #no_such_line_error_message, r15
        jmp     error

; Tokenize the line at tp
tokenize:
        mov.w   tp, r13                      ; R13 points to tokenized output
tokenize_01:
        mov.w   #kwtab, r14
        mov.b   #128, r12
        sub.w   #1, tp
tokenize_03:
        add.w   #1, tp
tokenize_05:
        mov.w   tp, r15
        mov.b   @r15, r4
        cmp.b   #0, r4
        jz      tokenize_end
        cmp.b   #' ', r4
        jz      tokenize_03
        cmp.b   #'"', r4
        jnz     tokenize_07
        add.w   #1, tp
tokenize_06:
        mov.b   r4, @r13
        add.w   #1, r13
        mov.b   @tp+, r4
        cmp.w   #0, r4
        jz      tokenize_end
        cmp.w   #'"', r4
        jnz     tokenize_06
        mov.b   r4, @r13
        add.w   #1, r13
        jmp     tokenize_01
tokenize_07:
        call    #uppercase
        cmp.b   @r14, r4
        jne     tokenize_20
tokenize_10:
        add.w   #1, r14
        add.w   #1, r15
        cmp.b   #0, @r14                     ; end of keyword and all matched?
        jz      tokenize_40
        cmp.b   #0, @r15                     ; end of input buffer
        jz      tokenize_20
        cmp.b   #'.', @r15                   ; keyword match forced
        jz      tokenize_30                  ; yep, replace it
        mov.b   @r15, r4
        call    #uppercase
        cmp.b   @r14, r4                     ; keyword and buffer continue to match?
        jz      tokenize_10                  ; ...yes, match next character
tokenize_20:
        mov.b   @r14+, r4                    ; skip to end of token
        cmp.b   #0, r4
        jne     tokenize_20
        add.w   #1, r12                      ; try next token
        cmp.b   #0, @r14                     ; end of token table
        jne     tokenize_05                  ; nope, try next token in table
        mov.b   @tp+, r4                     ; yes, copy character verbatim
        call    #uppercase
        mov.b   r4, @r13
        add.w   #1, r13
        jmp     tokenize_01
tokenize_30:
        add.w   #1, r15

; Contract the two tokens END WHILE to the single token WEND
tokenize_40:
        cmp.w   r13, &ramtop                 ; already have a token?
        jz      tokenize_50
        cmp.b   #T_EXIT, -1(r13)             ; EXIT
        jnz     tokenize_50
        cmp.b   #T_WHILE, r12                ; WHILE?
        jnz     tokenize_41
        mov.w   #T_EXIT_WHILE, r12           ; contract to EXIT WHILE
        sub.w   #1, r13                      ; write over previous token
tokenize_41:
        cmp.b   #T_FOR, r12                  ; FOR?
        jnz     tokenize_42
        mov.w   #T_EXIT_FOR, r12             ; contract to EXIT FOR
        sub.w   #1, r13                      ; write over previous token
tokenize_42:
        cmp.b   #T_REPEAT, r12               ; REPEAT?
        jnz     tokenize_50
        mov.w   #T_EXIT_REPEAT, r12          ; contract to EXIT REPEAT
        sub.w   #1, r13                      ; write over previous token
tokenize_50:
        mov.b   r12, @r13                    ; replace with tokenized form
        add.w   #1, r13
        mov.w   r15, tp
        jmp     tokenize_01
tokenize_end:
        mov.b   #0, @r13
tokenize_return:
        ret

; Check if r4 contains a character that ends a statement.
; Return with Z=1 if it does, else Z=0
end_of_statement:
        cmp.w   #0, r4
        jz      end_of_statement_return
        cmp.w   #':', r4
        jz      end_of_statement_return
        cmp.w   #T_TICK, r4
end_of_statement_return:
        ret

print_error_message:
        call    #print_message
        cmp.w   #0, lp
        jz      print_error_message_10
        mov.w   #in_line_message, r15
        call    #print_message
        mov.w   2(lp), r14
        mov.w   #0, r15
        call    #print_integer
print_error_message_10:
        mov.w   #'\n', r15
        jmp     print_char

print_bin:
        mov.w   r15, r13
        mov.w   r14, r12
        mov.w   #31, r14
print_bin_00:
	add.w   r12, r12
	addc.w  r13, r13
	mov.w   #'0', r15
	addc.w  #0, r15
	call    #print_char
	sub.w   #1, r14
	jnz     print_bin_00
	add.w   r12, r12
	addc.w  r13, r13
	mov.w   #'0', r15
	addc.w  #0, r15
	jmp     print_char
	
print_hex:
        mov.w   #0, r9                       ; not printed anything yet
        mov.w   r15, r13
        mov.w   r14, r12
        mov.w   #8, r14
print_hex_10:
        bit.w   #0x8000, r13
        addc.w  r12, r12
        addc.w  r13, r13
        bit.w   #0x8000, r13
        addc.w  r12, r12
        addc.w  r13, r13
        bit.w   #0x8000, r13
        addc.w  r12, r12
        addc.w  r13, r13
        bit.w   #0x8000, r13
        addc.w  r12, r12
        addc.w  r13, r13
        mov.w   r12, r15
        and.w   #0xf, r15
        cmp.b   #10, r15
        jnc     print_hex_20
        add.w   #'A'-'0'-10, r15
print_hex_20:
        add.w   #'0', r15
        cmp.w   #1, r14
        jz      print_hex_35
        cmp.w   #'0', r15
        jz      print_hex_30
        add.w   #1, r9
print_hex_30:
        cmp.w   #0, r9
        jz      print_hex_40
print_hex_35:
        call    #print_char
print_hex_40:
        sub.w   #1, r14
        jnz     print_hex_10        
print_hex_return:
        ret

print_integer:
        call    #convert_integer
        mov.w   #convert_buffer, r14
print_integer_10:
        mov.b   @r14+, r15
        cmp.w   #0, r15
        jz      print_hex_return
        call    #print_char
        jmp     print_integer_10

convert_integer:
        push.w  r11
        mov.w   #convert_buffer, r11
        mov.w   #0, r7
        mov.w   r15, r13
        mov.w   r14, r12
        cmp.w   #0, r13
        jge     convert_integer_00
        mov.b   #'-', @r11
        add.w   #1, r11
        xor.w   #-1, r12
        xor.w   #-1, r13
        add.w   #1, r12
        addc.w  #0, r13
convert_integer_00:
        mov.w   #pow10_table, r14
        cmp.w   #0, r15
        jnz     convert_integer_10
        mov.w   #pow10_table_16bit, r14
convert_integer_10:
        mov.w   #'0'-1, r15
convert_integer_20:
        add.w   #1, r15
        sub.w   @r14, r12
        subc.w  2(r14), r13
        jc      convert_integer_20
        add.w   @r14, r12
        addc.w  2(r14), r13
        cmp.w   #'0', r15
        jz      convert_integer_30
        add.w   #1, r7
convert_integer_30:
        cmp.w   #0, r7
        jz      convert_integer_40
        mov.b   r15, @r11
        add.w   #1, r11
convert_integer_40:
        add.w   #4, r14
        cmp.w   #pow10_table_end, r14
        jnz     convert_integer_10
        add.w   #'0', r12
        mov.b   r12, @r11
        add.w   #1, r11
        mov.b   #0, @r11
        pop.w   r11
        ret

print_message:
        push.w  r14
	mov.w   r15, r14
print_message_00:
        mov.b   @r14+, r15
        cmp.b   #0, r15
        jz      print_message_10
        call    #print_char
        jmp     print_message_00
print_message_10:
        pop.w   r14
        ret

; Read string into RAMTOP
get_string:
	mov.w   &ramtop, r14
get_string_00:
        call    #get_char
        cmp.b   #'\n', r15
        jz      get_string_10
        cmp.b   #'\b', r15
        jnz     get_string_05
        cmp.w   &ramtop, r14
        jz      get_string_00
        mov.b   #' ', r15
        call    #print_char
        mov.b   #'\b', r15
        call    #print_char
        sub.w   #1, r14
        jmp     get_string_00
get_string_05
        mov.b   r15, @r14
        add.w   #1, r14
        jmp     get_string_00
get_string_10:
        mov.b   #0, @r14
        ret

; write byte r4 to the flash at r10
flash_write_byte:
#ifdef SIMULATOR
        mov.b   r4, @r10
        ret
#else
        push.w  r15                          ; save working register
        mov.b   0(r10), r15                  ; get byte already in flash
        inv.b   r15                          ; can we overprogram the existing byte?
        and.b   r4, r15
        jnz     flash_write_byte_10          
        mov.w   #0xa500, &FCTL3
        mov.w   #0xa540, &FCTL1
        mov.b   r4, @r10
        mov.w   #0xa500, &FCTL1
        mov.w   #0xa510, &FCTL3
        pop.w   r15
        ret
flash_write_byte_10:
        pop.w   r15
        ret
#endif

flash_write_word:
#ifdef SIMULATOR
        mov.w   r4, @r10
        ret
#else
; Short-circuit if writing to RAM
        cmp.w   #0x1000, r10
        jc      flash_write_word_00
        mov.w   r4, @r10
        ret

flash_write_word_00:
        push.w  r15
        push.w  r14
        push.w  r13
        push.w  r12
        mov.w   0(r10), r15
        inv.w   r15
        and.w   r4, r15
        jnz     flash_write_word_10
        mov.w   #0xa500, &FCTL3
        mov.w   #0xa540, &FCTL1
        mov.w   r4, @r10
        mov.w   #0xa500, &FCTL1
        mov.w   #0xa510, &FCTL3
        jmp     flash_write_word_50

; Erase temp page
flash_write_word_10:
        mov.w   #0xa500, &FCTL3
        mov.w   #0xa502, &FCTL1
        mov.w   #0, &0xfc00
        mov.w   #0xa500, &FCTL1
        mov.w   #0xa510, &FCTL3

; Copy data to temp page, replacing byte
        mov.w   #512/2, r14
        mov.w   #0xfc00, r13
        mov.w   r10, r12
        bic.w   #0x1ff, r12

        mov.w   #0xa500, &FCTL3
        mov.w   #0xa540, &FCTL1

flash_write_word_20:
        cmp.w   r12, r10
        mov.w   @r12+, r15
        jnz     flash_write_word_30
        mov.w   r4, r15
flash_write_word_30:
        mov.w   r15, @r13
        add.w   #2, r13
        sub.w   #1, r14
        jnz     flash_write_word_20

        mov.w   #0xa500, &FCTL1
        mov.w   #0xa510, &FCTL3

; Erase original page
        mov.w   #0xa500, &FCTL3
        mov.w   #0xa502, &FCTL1
        bic.w   #0x1ff, r10
        mov.b   #0, @r10
        mov.w   #0xa500, &FCTL1
        mov.w   #0xa510, &FCTL3

; Copy working page to original page
        mov.w   #512/2, r13
        mov.w   #0xfc00, r12

        mov.w   #0xa500, &FCTL3
        mov.w   #0xa540, &FCTL1

flash_write_word_40:
        mov.w   @r12+, @r10
        add.w   #2, r10
        sub.w   #1, r13
        jne     flash_write_word_40

flash_write_word_50:
        pop.w   r12
        pop.w   r13
        pop.w   r14
        pop.w   r15
        ret
#endif

; Erase page pointed to by R10
flash_erase_page:
#ifdef SIMULATOR
        push.w  r11
        push.w  r10
        bic     #0x1ff, r10
        mov.w   #512/2, r11
flash_erase_page_00:
        mov.w   #-1, @r10
        add.w   #2, r10
        sub.w   #1, r11
        jnz     flash_erase_page_00
        pop.w   r10
        pop.w   r11
        ret
#else
        push.w  r11
        push.w  r10
        bic     #0x1ff, r10
        mov.w   #512/2, r11
flash_erase_page_00:
        cmp.w   #-1, @r10
        jnz     flash_erase_page_10
        add.w   #2, r10
        sub.w   #1, r11
        jnz     flash_erase_page_00
        pop.w   r10
        pop.w   r11
        ret
flash_erase_page_10:
        pop.w   r10
        mov.w   #0xa500, &FCTL3
        mov.w   #0xa502, &FCTL1
        mov.w   #0, @r10
        mov.w   #0xa500, &FCTL1
        mov.w   #0xa510, &FCTL3
        pop.w   r11
        ret
#endif

select_channel:
        call    #validate_channel
        mov.b   r15, &current_channel
        ret

validate_channel:
        cmp.w   #0, r15
        jnz     validate_channel_10
        mov.w   #1, r15
validate_channel_10:
        cmp.w   #4, r15
        jc      bad_channel_error
        ret

pow10_table:
        dc.l    1000000000
        dc.l    100000000
        dc.l    10000000
        dc.l    1000000
        dc.l    100000
pow10_table_16bit:
        dc.l    10000
        dc.l    1000
        dc.l    100
        dc.l    10
pow10_table_end:

; Print the character in R15
print_char:
        push.w  r14                          ; save R14 as we're going to use it
        mov.b   &current_channel, r14        ; get the current channel
        add.b   r14, r14                     ; compute index into channel print char table
        call    print_char_table-2(r14)      ; print the character in R15
        pop.w   r14                          ; restore R14
        ret                                  ; and return

print_at:
        push.w  r14
        mov.b   &current_channel, r14
        add.b   r14, r14
        call    print_at_table-2(r14)
        pop.w   r14
        ret

bad_channel_error:
        mov.w   #bad_channel_error_message, r15
        jmp     error

divide_by_zero_error:
        mov.w   #divide_by_zero_error_message, r15
        jmp     error

cant_resume_error:
        mov.w   #cant_resume_error_message, r15
        jmp     error

cant_continue_error:
        mov.w   #cant_continue_error_message, r15
        jmp     error

out_of_memory_error:
        mov.w   #out_of_memory_error_message, r15
        jmp     error

out_of_data_error:
        mov.w   #out_of_data_error_message, r15
        jmp     error

array_not_dimensioned_error:
        mov.w   #array_not_dimensioned_error_message, r15
        jmp     error

index_out_of_bounds_error:
        mov.w   #index_out_of_bounds_error_message, r15
        jmp     error

exit_not_in_loop_error:
        mov.w   #exit_not_in_loop_error_message, r15
        jmp     error

end_while_without_while_error:
        mov.w   #end_while_without_while_error_message, r15
        jmp     error

for_without_matching_next_error:
        mov.w   #for_without_matching_next_error_message, r15
        jmp     error

while_without_matching_end_while_error:
        mov.w   #while_without_matching_end_while_error_message, r15
        jmp     error

until_without_repeat_error:
        mov.w   #until_without_repeat_error_message, r15
        jmp     error

next_without_for_error:
        mov.w   #next_without_for_error_message, r15
        jmp     error

return_without_gosub_error:
        mov.w   #return_without_gosub_error_message, r15
        jmp     error

mismatched_parenthesis_error:
        mov.w   #mismatched_parenthesis_error_message, r15
        jmp     error

mismatched_quote_error:
        mov.w   #mismatched_quote_error_message, r15
        jmp     error

not_implemented_error:
        mov.w   #not_implemented_error_message, r15
        jmp     error

value_error:
        mov.w   #value_error_message, r15
        jmp     error

syntax_error:
        mov.w   #syntax_error_message, r15

; Entry point for a general error
error:
        mov.w   #0, &cont_tp                 ; obliterate continue context as we can't continue
        mov.w   #0, &cont_lp
        mov.b   #1, &current_channel
        call    #print_error_message
        jmp     ready

copyright_message:
        dc.b    "\n\nButterfly BASIC Interpreter v1.0 for the "
	dc.b    PLATFORM
	dc.b    ".\nCopyright (c) Paul Curtis 2003, 2004.\n\n"
#ifdef GRAPHICS
	dc.b    "GRAPHICS extension present.\n\n"
#endif
	dc.b	"Butterfly BASIC comes with ABSOLUTELY NO WARRANTY. This is free software\n"
	dc.b	"and you are welcome to redistribute it under certain conditions.  Refer to\n"
	dc.b	"the manual for extact details.\n\n", 0
divide_by_zero_error_message:
        dc.b    "*** DIVIDE BY ZERO ERROR", 0
not_implemented_error_message:
        dc.b    "*** NOT IMPLEMENTED", 0
cant_resume_error_message:
        dc.b    "*** CAN'T RESUME", 0
cant_continue_error_message:
        dc.b    "*** CAN'T CONTINUE", 0
out_of_data_error_message:
        dc.b    "*** OUT OF DATA", 0
out_of_memory_error_message:
        dc.b    "*** OUT OF MEMORY", 0
array_not_dimensioned_error_message:
        dc.b    "*** ARRAY NOT DIMENSIONED", 0
index_out_of_bounds_error_message:
        dc.b    "*** INDEX OUT OF BOUNDS", 0
syntax_error_message:
        dc.b    "*** SYNTAX ERROR", 0
value_error_message:
        dc.b    "*** VALUE ERROR", 0
no_such_line_error_message:
        dc.b    "*** NO SUCH LINE", 0
mismatched_parenthesis_error_message:
        dc.b    "*** MISMATCHED '('", 0
mismatched_quote_error_message:
        dc.b    "*** MISMATCHED '\"'", 0
exit_not_in_loop_error_message:
        dc.b    "*** EXIT NOT WITHIN LOOP", 0
end_while_without_while_error_message:
        dc.b    "*** END WHILE WITHOUT MATCHING WHILE", 0
until_without_repeat_error_message:
        dc.b    "*** UNTIL WITHOUT MATCHING REPEAT", 0
return_without_gosub_error_message:
        dc.b    "*** RETURN WITHOUT MATCHING GOSUB", 0
next_without_for_error_message:
        dc.b    "*** NEXT WITHOUT MATCHING FOR", 0
for_without_matching_next_error_message:
        dc.b    "*** FOR WITHOUT MATCHING NEXT", 0
while_without_matching_end_while_error_message:
        dc.b    "*** WHILE WITHOUT MATCHING END WHILE", 0
bad_channel_error_message:
        dc.b    "*** BAD CHANNEL ERROR", 0
break_in_message:
        dc.b    "*** BREAK", 0
bad_input_error_message:
        dc.b    "*** BAD INPUT\n", 0
in_line_message:
        dc.b    " IN LINE ", 0
ready_message:
        dc.b    ">", 0
ram_message:
        db      " RAM bytes free, ", 0
flash_message:
        db      " FLASH bytes free\n", 0
#ifdef GRAPHICS
graphics_message:
        dc.b    "Graphic extensions present.\n\n", 0
#endif

        even

        .vectors
        .keep
        org     RESET_VECTOR
        dw      reset
